# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
An implementation of a Bayes-like spam classifier.

Adapted from spambayes.
"""
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import *
from past.builtins import cmp
from builtins import range
from builtins import object
from past.utils import old_div

import math
from jw.util import configuration

LN2 = math.log(2)  # used frequently by chi-combining

def Chi2Q(x2, v):
    """
    Return probability(chisq >= x2, with v degrees of freedom).

    v must be even.
    """
    assert v & 1 == 0
    # XXX If x2 is very large, exp(-m) will underflow to 0.
    m = old_div(x2, 2.0)
    chiSum = term = math.exp(-m)
    for i in range(1, v//2):
        term *= old_div(m, i)
        chiSum += term
    # With small x2 and large v, accumulated roundoff error, plus error in
    # the platform exp(), can cause this to spill a few ULP above 1.0.  For
    # example, chi2Q(100, 300) on my box has sum == 1.0 + 2.0**-52 at this
    # point.  Returning a value even a teensy bit over 1.0 is no good.
    return min(chiSum, 1.0)

class WordInfo(object):
    # A WordInfo is created for each distinct word.  spamcount is the
    # number of trained spam msgs in which the word appears, and hamcount
    # the number of trained ham msgs.
    #
    # Invariant:  For use in a classifier database, at least one of
    # spamcount and hamcount must be non-zero.
    #
    # Important:  This is a tiny object.  Use of __slots__ is essential
    # to conserve memory.

    __slots__ = 'spamcount', 'hamcount'

    def __init__(self):
        self.__setstate__((0, 0))

    def __repr__(self):
        return "WordInfo(%d, %d)" % (self.spamcount, self.hamcount)

    def __getstate__(self):
        return self.spamcount, self.hamcount

    def __setstate__(self, t):
        self.spamcount, self.hamcount = t

class Classifier(object):

    def __init__(self, config=configuration.FromString('{}')):
        self.wordinfo = {}
        self.probcache = {}
        self.nspam = 0
        self.nham = 0
        self.useBigrams = config('use-bigrams', True)
        self.unknownWordStrength = config('unknown-word-strength', 0.45)
        self.unknownWordProbability = config('unknown-word-probability', 0.5)
        self.minimumProbabilityStrength = config('minimum-probability-strength', 0.1)
        self.maxDiscriminators = config('max-discriminators', 150)

    def spamProbability(self, wordstream, evidence=False):
        """
        Return best-guess probability that wordstream is spam.

        wordstream is an iterable object producing words.
        The return value is a float in [0.0, 1.0].

        If optional arg evidence is True, the return value is a pair
            probability, evidence
        where evidence is a list of (word, probability) pairs.

        Across vectors of length n, containing random uniformly-distributed
        probabilities, -2*sum(ln(p_i)) follows the chi-squared distribution
        with 2*n degrees of freedom.  This has been proven (in some
        appropriate sense) to be the most sensitive possible test for
        rejecting the hypothesis that a vector of probabilities is uniformly
        distributed.  Gary Robinson's original scheme was monotonic *with*
        this test, but skipped the details.  Turns out that getting closer
        to the theoretical roots gives a much sharper classification, with
        a very small (in # of msgs), but also very broad (in range of scores),
        "middle ground", where most of the mistakes live.  In particular,
        this scheme seems immune to all forms of "cancellation disease":  if
        there are many strong ham *and* spam clues, this reliably scores
        close to 0.5.  Most other schemes are extremely certain then -- and
        often wrong.
        """

        # We compute two chi-squared statistics, one for ham and one for
        # spam.  The sum-of-the-logs business is more sensitive to probs
        # near 0 than to probs near 1, so the spam measure uses 1-p (so
        # that high-spamprob words have greatest effect), and the ham
        # measure uses p directly (so that lo-spamprob words have greatest
        # effect).
        #
        # For optimization, sum-of-logs == log-of-product, and f.p.
        # multiplication is a lot cheaper than calling ln().  It's easy
        # to underflow to 0.0, though, so we simulate unbounded dynamic
        # range via frexp.  The real product H = this H * 2**Hexp, and
        # likewise the real product S = this S * 2**Sexp.
        h = s = 1.0
        hExp = sExp = 0
        clues = self._clues(wordstream)
        for prob, word, record in clues:
            s *= 1.0 - prob
            h *= prob
            if s < 1e-200:  # prevent underflow
                s, e = math.frexp(s)
                sExp += e
            if h < 1e-200:  # prevent underflow
                h, e = math.frexp(h)
                hExp += e
        # Compute the natural log of the product = sum of the logs:
        # ln(x * 2**i) = ln(x) + i * ln(2).
        s = math.log(s) + sExp * LN2
        h = math.log(h) + hExp * LN2
        n = len(clues)
        if n:
            s = 1.0 - Chi2Q(-2.0 * s, 2 * n)
            h = 1.0 - Chi2Q(-2.0 * h, 2 * n)
            # How to combine these into a single spam score?  We originally
            # used (S-H)/(S+H) scaled into [0., 1.], which equals S/(S+H).  A
            # systematic problem is that we could end up being near-certain
            # a thing was (for example) spam, even if S was small, provided
            # that H was much smaller.
            # Rob Hooft stared at these problems and invented the measure
            # we use now, the simpler S-H, scaled into [0., 1.].
            prob = old_div((s - h + 1.0), 2.0)
        else:
            prob = 0.5
        if evidence:
            clues = [(w, p) for p, w, _r in clues]
            clues.sort(lambda a, b: cmp(a[1], b[1]))
            clues.insert(0, ('*S*', s))
            clues.insert(0, ('*H*', h))
            return prob, clues
        else:
            return prob

    def learn(self, wordstream, isSpam):
        """
        Teach the classifier by example.

        wordstream is a word stream representing a message.  If isSpam is
        True, you're telling the classifier this message is definitely spam,
        else that it's definitely not spam.
        """
        if self.useBigrams:
            wordstream = self._enhanceWordstream(wordstream)
        self._addMsg(wordstream, isSpam)

    def unlearn(self, wordstream, is_spam):
        """In case of pilot error, call unlearn ASAP after screwing up.

        Pass the same arguments you passed to learn().
        """
        if self.useBigrams:
            wordstream = self._enhanceWordstream(wordstream)
        self._removeMsg(wordstream, is_spam)

    def probability(self, record):
        """
        Compute, store, and return prob(msg is spam | msg contains word).

        This is the Graham calculation, but stripped of biases, and
        stripped of clamping into 0.01 thru 0.99.  The Bayesian
        adjustment following keeps them in a sane range, and one
        that naturally grows the more evidence there is to back up
        a probability.
        """
        spamcount = record.spamcount
        hamcount = record.hamcount
        # Try the cache first
        try:
            return self.probcache[spamcount][hamcount]
        except KeyError:
            pass
        nham = float(self.nham or 1)
        nspam = float(self.nspam or 1)
        assert hamcount <= nham, "Token seen in more ham than ham trained."
        hamratio = old_div(hamcount, nham)
        assert spamcount <= nspam, "Token seen in more spam than spam trained."
        spamratio = old_div(spamcount, nspam)
        prob = old_div(spamratio, (hamratio + spamratio))
        S = self.unknownWordStrength
        StimesX = S * self.unknownWordProbability
        # Now do Robinson's Bayesian adjustment.
        #
        #         s*x + n*p(w)
        # f(w) = --------------
        #           s + n
        #
        # I find this easier to reason about like so (equivalent when
        # s != 0):
        #
        #        x - p
        #  p +  -------
        #       1 + n/s
        #
        # IOW, it moves p a fraction of the distance from p to x, and
        # less so the larger n is, or the smaller s is.
        n = hamcount + spamcount
        prob = old_div((StimesX + n * prob), (S + n))
        # Update the cache
        try:
            self.probcache[spamcount][hamcount] = prob
        except KeyError:
            self.probcache[spamcount] = {hamcount: prob}
        return prob

    def _addMsg(self, wordstream, isSpam):
        self.probcache = {}  # nuke the prob cache
        if isSpam:
            self.nspam += 1
        else:
            self.nham += 1

        for word in set(wordstream):
            record = self._wordInfo(word)
            if record is None:
                record = WordInfo()

            if isSpam:
                record.spamcount += 1
            else:
                record.hamcount += 1

            self._setWordInfo(word, record)

    def _removeMsg(self, wordstream, is_spam):
        # NOTE:  Graham's scheme had a strange asymmetry:  when a word appeared
        # n>1 times in a single message, training added n to the word's hamcount
        # or spamcount, but predicting scored words only once.  Tests showed
        # that adding only 1 in training, or scoring more than once when
        # predicting, hurt under the Graham scheme.
        # This isn't so under Robinson's scheme, though:  results improve
        # if training also counts a word only once.  The mean ham score decreases
        # significantly and consistently, ham score variance decreases likewise,
        # mean spam score decreases (but less than mean ham score, so the spread
        # increases), and spam score variance increases.
        # I (Tim) speculate that adding n times under the Graham scheme helped
        # because it acted against the various ham biases, giving frequently
        # repeated spam words (like "Viagra") a quick ramp-up in spamprob; else,
        # adding only once in training, a word like that was simply ignored until
        # it appeared in 5 distinct training spams.  Without the ham-favoring
        # biases, though, and never ignoring words, counting n times introduces
        # a subtle and unhelpful bias.
        # There does appear to be some useful info in how many times a word
        # appears in a msg, but distorting spamprob doesn't appear a correct way
        # to exploit it.
        self.probcache = {}  # nuke the prob cache
        if is_spam:
            if self.nspam <= 0:
                raise ValueError("spam count would go negative!")
            self.nspam -= 1
        else:
            if self.nham <= 0:
                raise ValueError("non-spam count would go negative!")
            self.nham -= 1
        for word in set(wordstream):
            record = self._wordInfo(word)
            if record is not None:
                if is_spam:
                    if record.spamcount > 0:
                        record.spamcount -= 1
                else:
                    if record.hamcount > 0:
                        record.hamcount -= 1
                if record.hamcount == 0 == record.spamcount:
                    self._deleteWordInfo(word)
                else:
                    self._setWordInfo(word, record)

    def _clues(self, wordstream):
        """
        Return list of (prob, word, record) triples, sorted by increasing
        prob.  "word" is a token from wordstream; "prob" is its spamprob (a
        float in 0.0 through 1.0); and "record" is word's associated
        WordInfo record if word is in the training database, or None if it's
        not.  No more than max_discriminators items are returned, and have
        the strongest (farthest from 0.5) spamprobs of all tokens in wordstream.
        Tokens with spamprobs less than minimum_prob_strength away from 0.5
        aren't returned.
        """
        mindist = self.minimumProbabilityStrength
        if self.useBigrams:
            # This scheme mixes single tokens with pairs of adjacent tokens.
            # wordstream is "tiled" into non-overlapping unigrams and
            # bigrams.  Non-overlap is important to prevent a single original
            # token from contributing to more than one spamprob returned
            # (systematic correlation probably isn't a good thing).
            #
            # First fill list raw with
            #     (distance, prob, word, record), indices
            # pairs, one for each unigram and bigram in wordstream.
            # indices is a tuple containing the indices (0-based relative to
            # the start of wordstream) of the tokens that went into word.
            # indices is a 1-tuple for an original token, and a 2-tuple for
            # a synthesized bigram token.  The indices are needed to detect
            # overlap later.
            raw = []
            push = raw.append
            pair = None
            # Keep track of which tokens we've already seen.
            # Don't use a set here!  This is an innermost loop, so speed is
            # important here (direct dict fiddling is much quicker than
            # invoking Python-level set methods; in Python 2.4 that will
            # change).
            seen = {pair: 1}  # so the bigram token is skipped on 1st loop trip
            last_token = None
            for i, token in enumerate(wordstream):
                if i:
                    # not the 1st loop trip, so there is a preceding token
                    # This string interpolation must match the one in
                    # _enhance_wordstream().
                    pair = "bi:%s %s" % (last_token, token)
                last_token = token
                for clue, indices in (token, (i,)), (pair, (i - 1, i)):
                    if clue not in seen:  # as always, skip duplicates
                        seen[clue] = 1
                        tup = self._wordDistance(clue)
                        if tup[0] >= mindist:
                            push((tup, indices))
            # Sort raw, strongest to weakest spamprob.
            raw.sort(reverse=True)
            # Fill clues with the strongest non-overlapping clues.
            clues = []
            push = clues.append
            # Keep track of which indices have already contributed to a
            # clue in clues.
            seen = {}
            for tup, indices in raw:
                overlap = [i for i in indices if i in seen]
                if not overlap:  # no overlap with anything already in clues
                    for i in indices:
                        seen[i] = 1
                    push(tup)
            # Leave sorted from smallest to largest spamprob.
            clues.reverse()
        else:
            # The all-unigram scheme just scores the tokens as-is.  A set()
            # is used to weed out duplicates at high speed.
            clues = []
            push = clues.append
            for word in set(wordstream):
                tup = self._wordDistance(word)
                if tup[0] >= mindist:
                    push(tup)
            clues.sort()
        if len(clues) > self.maxDiscriminators:
            del clues[0: -self.maxDiscriminators]
        # Return (prob, word, record).
        return [t[1:] for t in clues]

    def _wordDistance(self, word):
        record = self._wordInfo(word)
        if record is None:
            prob = self.unknownWordProbability
        else:
            prob = self.probability(record)
        distance = abs(prob - 0.5)
        return distance, prob, word, record

    def _wordInfo(self, word):
        return self.wordinfo.get(word)

    def _setWordInfo(self, word, record):
        self.wordinfo[word] = record

    def _deleteWordInfo(self, word):
        del self.wordinfo[word]

    @staticmethod
    def _enhanceWordstream(wordstream):
        """Add bigrams to the wordstream.

        For example, a b c -> a b "a b" c "b c"

        Note that these are *token* bigrams, and not *word* bigrams - i.e.
        'synthetic' tokens get bigram'ed, too.

        The bigram token is simply "bi:unigram1 unigram2" - a space should
        be sufficient as a separator, since spaces aren't in any other
        tokens, apart from 'synthetic' ones.  The "bi:" prefix is added
        to avoid conflict with tokens we generate (like "subject: word",
        which could be "word" in a subject, or a bigram of "subject:" and
        "word").

        If the "Classifier":"use_bigrams" option is removed, this function
        can be removed, too.
        """
        last = None
        for token in wordstream:
            yield token
            if last:
                # This string interpolation must match the one in
                # _getclues().
                yield "bi:%s %s" % (last, token)
            last = token



"""
URL filter (spamvertized mail)
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import *
import email
import logging

import mock
from nose.tools import ok_
from gevent import socket
from mailfilter import filters

logging.basicConfig(level=logging.DEBUG)

def test10_positive_url():
    """Positive for spam URL"""
    imap = mock.Mock()
    tfilter = filters.UrlFilter(imap, {'filter': 'url', 'action': 'delete'}, {})
    message = email.message_from_file(open('test/url/spam_positive.msg'))
    # Pretend gethostbyname HAS an entry about the IP in the mail
    with mock.patch('mailfilter.filters.gethostbyname', return_value='127.0.0.10') as f:
        logging.info('---')
        ok_(not tfilter.check(1, '', message), 'UrlFilter.check() failed')
        ok_(f.called, 'gethostbyname() never called')

def test20_negative_url():
    """Negative for spam URL"""
    imap = mock.Mock()
    tfilter = filters.UrlFilter(imap, {'filter': 'url', 'action': 'delete'}, {})
    message = email.message_from_file(open('test/url/spam_negative.msg'))
    # Pretend gethostbyname has no entry about the IP in the mail
    with mock.patch('mailfilter.filters.gethostbyname', side_effect=socket.gaierror) as f:
        ok_(tfilter.check(1, '', message), 'UrlFilter.check() failed')
        ok_(f.called, 'gethostbyname() never called')



# -*- coding: UTF-8 -*-
"""
Test header filter
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import io
from future import standard_library
standard_library.install_aliases()
from builtins import *
from nose.tools import ok_
import email
import mock
from mailfilter import filters
import pdb

def test10_positive_exact():
    """Positive for exact 'test mailing'"""
    imap = mock.Mock()
    tfilter = filters.HeaderFilter(imap, {'filter': 'header', 'part': 'subject', 'check': "has 'test mailing'", 'action': 'delete'}, {})
    message = email.message_from_file(open('test/header/positive_exact.msg'))
    ok_(not tfilter.check(1, '', message), "HeaderFilter.check() failed")

def test11_positive__encoded_exact():
    """Positive for exact encoded from ''"""
    imap = mock.Mock()
    tfilter = filters.HeaderFilter(imap, {'filter': 'header', 'part': 'subject', 'check': "is 'Heizen Sie für immer gratis!'", 'action': 'delete'}, {})
    message = email.message_from_file(open('test/header/encoded.msg'))
    ok_(not tfilter.check(1, '', message), "HeaderFilter.check() failed")

def test20_negative_exact():
    """Negative for 'test  mailing'"""
    imap = mock.Mock()
    tfilter = filters.HeaderFilter(imap, {'filter': 'header', 'part': 'subject', 'check': "has 'test mailing'", 'action': 'delete'}, {})
    message = email.message_from_file(open('test/header/positive_word.msg'))
    ok_(tfilter.check(1, '', message), "HeaderFilter.check() failed")

def test30_positive_upper():
    """Positive for 'TEST MAILING'"""
    imap = mock.Mock()
    tfilter = filters.HeaderFilter(imap, {'filter': 'header', 'part': 'subject', 'check': "has 'test mailing'", 'action': 'delete'}, {})
    message = email.message_from_file(open('test/header/positive_upper.msg'))
    ok_(not tfilter.check(1, '', message), "HeaderFilter.check() failed")

def test40_positive_word():
    """Positive for 'test mailing'"""
    imap = mock.Mock()
    tfilter = filters.HeaderFilter(imap, {'filter': 'header', 'part': 'subject', 'check': "has test mailing", 'action': 'delete'}, {})
    message = email.message_from_file(open('test/header/positive_word.msg'))
    ok_(not tfilter.check(1, '', message), "HeaderFilter.check() failed")

def test50_negative():
    """Negative for 'test mailing'"""
    imap = mock.Mock()
    tfilter = filters.HeaderFilter(imap, {'filter': 'header', 'part': 'subject', 'check': "has 'test mailing'", 'action': 'delete'}, {})
    message = email.message_from_file(open('test/header/negative.msg'))
    ok_(tfilter.check(1, '', message), "HeaderFilter.check() failed")



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Content checks
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from future import standard_library
from jw.util.python3hell import Bytes2Str

standard_library.install_aliases()
from builtins import *

import re

import logging
from .logging2 import *

class Check(object):
    """
    Check base class
    """

class Has(Check):
    """
    Check for one or more text occurrences

    Checks whether any (at least one) of the arguments appears in the text
    """

    def __init__(self, filterDef, args):
        """
        Create Has object

        """
        self.args = tuple(re.compile(a.strip().lower()) for a in args)
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))

    def check(self, id, content):
        """
        Run check

        :param id: Message ID
        :type id: int
        :param content: content to check
        :type content: str
        :return: True if positive
        :rtype: bool
        """
        lcontent = Bytes2Str(content).lower()
        result = any(arg.search(lcontent) for arg in self.args)
        if result:
            self.log.debug('Message %d has some of %s', id, self.args)
        return result

class HasNo(Check):
    def __init__(self, filterDef, args):
        """
        Create HasNo object

        Checks whether none of the arguments appear in the text
        """
        self.args = tuple(re.compile(a.strip().lower()) for a in args)
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))

    def check(self, id, content):
        """
        Run check

        :param id: Message ID
        :type id: int
        :param content: content to check
        :type content: str
        :return: True if positive
        :rtype: bool
        """
        lcontent = Bytes2Str(content).lower()
        result = all(not arg.search(lcontent) for arg in self.args)
        if result:
            self.log.debug('Message %d has some of %s', id, self.args)
        return result

class HasAll(Check):
    def __init__(self, filterDef, args):
        """
        Create HasAll object

        Checks whether all arguments appear in the text
        """
        self.args = tuple(re.compile(a.strip().lower()) for a in args)
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))

    def check(self, id, content):
        """
        Run check

        :param id: Message ID
        :type id: int
        :param content: content to check
        :type content: str
        :return: True if positive
        :rtype: bool
        """
        lcontent = Bytes2Str(content).lower()
        result = all(arg.search(lcontent) for arg in self.args)
        if result:
            self.log.debug('Message %d has some of %s', id, self.args)
        return result

class Is(Check):
    def __init__(self, filterDef, args):
        """
        Create Has object

        Checks whether text is exactly the arguments
        """
        assert len(args) == 1, 'is check accepts only one argument, got %s' % repr(args)
        self.args = re.compile(Bytes2Str(args[0]).strip().lower(), re.DOTALL | re.MULTILINE)
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))

    def check(self, id, content):
        """
        Run check

        :param id: Message ID
        :type id: int
        :param content: content to check
        :type content: str
        :return: True if positive
        :rtype: bool
        """
        s = Bytes2Str(content).strip().lower()
        result = self.args.match(s) is not None
        if result:
            self.log.debug('Message %d match %s', id, self.args.pattern)
        else:
            self.log.log(DEBUG2, 'Message %d %s does not match %s', id, s, self.args.pattern)
        return result

class IsNot(Check):
    def __init__(self, filterDef, args):
        """
        Create Has object

        Checks whether text is not the arguments
        """
        assert len(args) == 1, 'is-not check accepts only one argument, got %s' % repr(args)
        self.args = re.compile(args[0].strip().lower(), re.DOTALL | re.MULTILINE)
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))

    def check(self, id, content):
        """
        Run check

        :param id: Message ID
        :type id: int
        :param content: content to check
        :type content: str
        :return: True if positive
        :rtype: bool
        """
        result = self.args.match(Bytes2Str(content).strip().lower()) is None
        if result:
            self.log.debug('Message %d non-match %s', id, self.args.pattern)
        return result

FILTER_FUNCTIONS = {
    'is': Is,
    'is-not': IsNot,
    'has': Has,
    'has-no': HasNo,
    'has-all': HasAll
}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrapper for IMAPClient class
"""

from __future__ import absolute_import, print_function, unicode_literals
from __future__ import division
from future import standard_library
from .logging2 import *

standard_library.install_aliases()
from builtins import *
from time import sleep, time
from jw.util.types import Args2Str
from imapclient import IMAPClient

import logging

IMAP_RETRIES = 10

class IMAP(IMAPClient):
    """
    IMAP access
    """

    def __init__(
            self, host, username, password, *args, **kwargs
    ):
        self.username = username
        self.password = password
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))
        super(IMAP, self).__init__(host, *args, **kwargs)
        self.login(username, password)

    def __call__(self, name, *args, **kwargs):
        stime = 1
        t1 = time()
        for i in range(IMAP_RETRIES):
            try:
                result = getattr(self, name)(*args, **kwargs)
                self.log.log(DEBUG2, 'IMAP: %s(%s) -> %s', name, Args2Str(args, kwargs), result)
                break
            except self.AbortError as e:
                self.log.warning(
                    'From %s(%s) in %s: %s. Next try in %d sec.', name, Args2Str(*args, **kwargs), self.host, e, stime
                )
                sleep(stime)
                stime *= 2
                try:
                    self._imap.shutdown()
                except Exception as e2:
                    self.log.warning('From self.shutdown() in %s: %s', self.host, e2)
                try:
                    self._imap = self._create_IMAP4()
                except Exception as e2:
                    self.log.warning('From self._create_IMAP4() in %s: %s', self.host, e2)
                    continue
                try:
                    self.login(self.username, self.password)
                except Exception as e2:
                    self.log.warning('From login("%s", "***") in %s: %s', self.username, self.host, e2)
                    continue
        else:
            self.log.critical(
                '%s(%s) in %s: Error after %s tries. Giving up.', name, Args2Str(*args, **kwargs), self.host, i + 1
            )
            return
        if i:
            t2 = time()
            self.log.warning(
                '%s(%s) in %s: Successful after %s tries in %.1f sec.',
                name,
                Args2Str(*args, **kwargs),
                self.host,
                i,
                t2 - t1
            )
        return result



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Account
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from functools import partial
from pprint import pprint
from future import standard_library
from jw.util.retry import retry

standard_library.install_aliases()
from itertools import takewhile, chain
import email

from builtins import str
from builtins import object
import gevent
from . import filters
from . import imap

import logging
from jw.util.extension import LoadClass
import ssl
from .util import Flatten, D, DecodeHeader

#: Default IDLE timeout
IDLE_TIMEOUT = 180

class UndefinedFilter(RuntimeError):
    def __init__(self, error):
        RuntimeError.__init__(self, 'Undefined filter type: ' + error)

class Folder(object):

    def __init__(self, name, folders, config, globalConfig=None):
        """

        :param name: Mail filter name from config
        :type name: str
        :param config: config
        :type config: dict
        """
        assert isinstance(folders, list), 'folders must be a list'
        # Parameters
        self.name = name
        self.config = config
        self.globalConfig = globalConfig or {}
        self.timeout = config.get('time-out', globalConfig.get('time-out'))
        # Logging
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))
        # Miscellaneous setup
        self.imap = None
        self.seen = {}
        if not self._startImap():
            raise RuntimeError('Account %s not startable' % (self.name))
        # Setup folders and filters
        self.folders = [
            (name, tuple(self._filterFromString(f) for f in Flatten(filters))) for name, filters in folders
        ]
        for folder, filter in self.folders:
            self.seen[folder] = set()
            if any(f.needsBody for f in filter):
                # Full body required
                self.fetchPart = self.messagePart = 'RFC822'
            else:
                # Only headers required
                self.fetchPart = 'BODY.PEEK[HEADER]'
                self.messagePart = 'BODY[HEADER]'
        # Log
        self.log.info(
            'Account %s, folders %s: %s (%s)',
            self.name,
            ', '.join(str(f) for f in self.folders),
            self.config['address'],
            ('No SSL', 'SSL')[self.config.get('ssl', False)]
        )

    def run(self):
        """
        Mail filter

        """
        self.separator = self.imap('namespace')[0][0][1]
        idleOk = 'IDLE' in self.imap('capabilities')
        idleUsed = self.config.get('idle', idleOk)
        if idleUsed and idleUsed != idleOk:
            raise RuntimeError('Server does not support IDLE: ' + self.name)
        self._check('ALL')
        if idleUsed:
            self.imap('idle')
        while True:
            if idleUsed:
                idleResult = self.imap('idle_check', self.timeout)
                if idleResult and idleResult[1]:
                    self.log.debug('%s: %s', self.name, idleResult)
                    if idleUsed:
                        self.imap('idle_done')
                    try:
                        self._check()
                    except Exception as e:
                        self.log.exception('Uncaught exception:')
                    if idleUsed:
                        self.imap('idle')
            else:
                gevent.sleep(self.timeout or IDLE_TIMEOUT)
                try:
                    self._check()
                except Exception as e:
                    self.log.exception('Uncaught exception:')
        self.log.critical('%s stopped', self.name)
        self.imap = None

    def _check(self, search='ALL'):
        """
        Check messages
        """
        # Check all folders in account
        for folder, filters in self.folders:
            self.log.debug('Checking folder %s in %s', folder, self.name)
            # Try to select folder
            try:
                self.imap('select_folder', folder.replace('/', self.separator))
            except Exception as e:
                self.log.critical('From self.imap.select_folder("%s"): %s', folder, e)
                continue
            # Determine messages to scan
            idlist = set(self.imap('search', search))
            seen = self.seen[folder]
            todo = idlist - seen
            #self.log.debug('Seen: %s', seen)
            self.log.debug('To do: %s', todo)
            if not todo:
                continue
            mlist = self.imap('fetch', todo, self.fetchPart)
            # Scan messages
            for m, msg in list(mlist.items()):
                self.log.debug('Checking message %s', m)
                # Get message from IMAP
                try:
                    message = email.message_from_string(msg[self.messagePart])
                except KeyError as e:
                    self.log.critical('%s part not found in message %s:\n%s', self.messagePart, msg, e)
                    continue
                subject = DecodeHeader(message['Subject'])
                # Do checks until one of them returns 'stop'
                # [Calls to filter.check() also trigger the actions]
                checks = list(
                    takewhile(
                        lambda item: item[1] != 'stop',
                        ((filter.filterDef['filter'], filter.check(m, self, message)) for filter in filters)
                    )
                )
                result = all(c[1] for c in checks)
                self.log.info('%s %s(%s): "%s" is %s %s', m, self.name, folder, subject, ('spam', 'ham')[result], checks)
                # Add processed message
                seen.add(m)
            self.imap('expunge')

    def _startImap(self):
        # If connection exists, shut it down
        if self.imap is not None:
            try:
                self.imap.shutdown()
            except Exception as e:
                self.log.critical('From self.imap.shutdown() in %s: %s', self.name, e)
        # Set up SSL context
        context = ssl.create_default_context()
        context.check_hostname = False  # TODO: make configurable
        context.verify_mode = ssl.CERT_NONE  # TODO: make configurable
        # Create connection
        self.imap = retry(
            imap.IMAP,
            self.config['address'],
            self.config['username'],
            self.config['password'],
            ssl=self.config.get('ssl', False),
            ssl_context=context,
            logFailure_=partial(self.log.critical, 'imap.IMAP(%s): %s', self.name)
        )
        self.log.debug('Capabilites: %s', self.imap.capabilities())
        # Try to do STARTTLS
        if self.config.get('starttls', False):
            self.log.info('STARTTLS')
            try:
                self.imap._command_and_check('STARTTLS')
            except:
                self.log.exception('From STARTTLS command:')
        return True

    def _filterFromString(self, filterDef):
        """
        Create a Filter object from a definition

        :param filterDef:
        :type filterDef: dict
        :return: filter object
        :rtype: filters.Filter
        :raise filter.FilterError: if filter name is undefined
        """
        try:
            filterClass = LoadClass('jw.mailfilter.filter', filterDef['filter'], type_=filters.Filter)
        except KeyError:
            raise UndefinedFilter(str(filterDef))
        except Exception as e:
            raise filters.FilterError('Account definition error %s: %s' % (repr(e), repr(filterDef)))
        return filterClass(self.imap, filterDef, self.globalConfig)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
The filter module provides various filters for sorting out spam
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from future import standard_library

from .util import FromRe, DecodeHeader

standard_library.install_aliases()
from socket import gaierror
from itertools import chain
import logging
import re
import shlex
from urllib.parse import urlparse
import bs4
from gevent.socket import gethostbyname
import pyzor.client
from pyzor.digest import DataDigester
from jw.util.memoize import memoizeFor
from jw.util.extension import LoadClass, ResourceNotFound
from jw.util.split import Splitter

from . import action
from . import content
from .logging2 import *

SPAM_URL_CACHE_TIME = 3600
SPLIT_RE = re.compile(r'"[^"]*(?:"|$)|\S+')

Log = logging.getLogger(__name__)

@memoizeFor(SPAM_URL_CACHE_TIME)
def CheckSpamUrl(url):
    """
    Check a URL whether it's spamvertised

    :param url: URL
    :type url: str
    :return: True if spamvertised
    :rtype : bool
    """
    parts = url.split('.')
    for i in range(2, 4):
        if len(parts) >= i:
            cpart = '.'.join(parts[-i:])
            if cpart:
                try:
                    ip = gethostbyname(cpart + '.multi.surbl.org')
                    Log.info('URL %s positive', cpart)
                    if ip.startswith('127.'):
                        return True
                    else:
                        Log.warning('Weird result from multi.surbl.org: %s', ip)
                except gaierror:
                    Log.debug('URL %s ok', cpart)
                except Exception as e:
                    Log.critical('gethostbyname of %s (%s) yielded %s', repr(cpart), repr(parts), e)
    return False

def GetPayload(part):
    payload = part.get_payload(decode=True)
    if payload is None:
        payload = ''
    else:
        if isinstance(payload, bytes):
            payload = payload.decode(part.get_content_charset('latin1'), errors='replace')
        elif not isinstance(payload, str):
            Log.error('Payload %s is %s', repr(payload)[:256])
            payload = ''
    return payload

class SpamActionError(RuntimeError):
    """
    Error on spam action parsing
    """

class FilterError(RuntimeError):
    """
    Error parsing filter arguments
    """

class Filter(object):
    """
    Filter base

    """
    def __init__(self, imap, filterDef, globalConfig):
        """
        Create Filter object

        :param imap: IMAP client
        :type imap: imapclient.IMAPClient
        :param filterDef:
        :type filterDef: dict
        :raise SpamActionError:

        Needs to be called from derived classes
        """
        # General setup
        self._needsBody = False
        # Copy parameters
        self.imap = imap
        self.filterDef = filterDef
        self.globalConfig = globalConfig
        # Setup action
        self.spamAction = []
        actions = filterDef['action']
        if not isinstance(actions, list):
            actions = actions,
        for a in actions:
            lexer = shlex.shlex(a, posix=True)
            lexer.whitespace_split = True
            spam = list(lexer)
            try:
                actionClass = LoadClass('jw.mailfilter.action', spam[0], type_=action.Action)
                action_ = actionClass(imap, spam[1:], self.globalConfig)
                self.spamAction.append(action_)
                self._needsBody = self._needsBody or action_.needsBody
            except ResourceNotFound:
                raise SpamActionError('Undefined spam action %s in %s' % (spam[0], filterDef))

    def check(self, id, folder, message):
        """
        Run filter

        :param id:
        :type id: int
        :param message:
        :type message: email.email
        :return: True if good, False if bad
        :rtype: bool
        """
        raise NotImplementedError('check')

    def spam(self, id, folder, message):
        """
        Run spam action

        :param id:
        :type id: int
        :param message:
        :type message: email.message
        """
        for a in self.spamAction:
            a.run(id, folder, message)

    @property
    def needsBody(self):
        """
        Get body requirements

        :return: True if filter needs examination of BODY
        :rtype: bool
        """
        return self._needsBody

    @staticmethod
    def extract(message):
        """
        Extract payload

        :param message:
        :type message:
        :return: body
        :rtype: unicode
        """
        if message.is_multipart():
            return ''.join(BodyFilter.extract(m) for m in message.get_payload(decode=True))
        else:
            return message.get_payload(decode=True) or message.get_payload(decode=True)

class All(Filter):
    def __init__(self, imap, filterDef, globalConfig):
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))
        super(All, self).__init__(imap, filterDef, globalConfig)

    def check(self, msgid, folder, message):
        """
        Run filter

        :param msgid:
        :param message:
        :return: False
        :rtype: bool

        Returns False for all mail
        """
        self.spam(msgid, folder, message)
        return False

class RblFilter(Filter):
    """
    RBL filter
    """
    def __init__(self, imap, filterDef, globalConfig):
        error = False
        try:
            self.rbl = filterDef['org']
            regex = filterDef.get('positiv')
            self.re = re.compile(regex) if regex else None
        except:
            error = True
        if error:
            raise FilterError(
                'rbl filter syntax is: {filter: rbl, org: domain, [positiv: regex], spam: action}, got "%s"' % repr(filterDef)
            )
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))
        super(RblFilter, self).__init__(imap, filterDef, globalConfig)

    def check(self, msgid, folder, message):
        """
        Run filter

        :param msgid:
        :type msgid: int
        :param message:
        :type message: email.email
        :return: True if ok
        :rtype: bool
        """
        received = message.get_all('received')
        if received:
            ipv4list = list(chain.from_iterable(FromRe.findall(i) for i in received if i.startswith('from ')))
            if not ipv4list:
                ipv6list = list(
                    chain.from_iterable(
                        re.findall(r'[\[(](\d+\.\d+\.\d+\.\d+)[\])]', i) for i in message.get_all('received')
                    )
                )
                self.log.warning('No Received-headers with usable IPs in message')
                return False
            else:
                ip = ipv4list[0][1]
                revip = '.'.join(reversed(ip.split('.')))
                dname = '{}.{}'.format(revip, self.rbl)
                try:
                    result = gethostbyname(dname)
                except:
                    # Not spam
                    pass
                else:
                    # Prevent positive (spam) result only if a result regex is not matched
                    if self.re and self.re.match(result) is None:
                        return True
                    self.spam(msgid, folder, message)
                    self.log.info('Message %d from spammer IP %s', msgid, ip)
                    return False
        return True

class ContentFilter(Filter):
    """
    Content filter
    """

    def __init__(self, imap, filterDef, globalConfig):
        """
        Create ContentFilter object

        :param imap: IMAP client
        :type imap: imapclient.IMAPClient
        :param filterDef: filter definition
        :type filterDef: dict
        """
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))
        # Setup check
        args = Splitter(filterDef['check']).split()
        if args[0] not in content.FILTER_FUNCTIONS:
            raise FilterError('No such content check: ' + filterDef['check'])
        else:
            self.filterFunction = content.FILTER_FUNCTIONS[args[0]](filterDef, args[1:])
        super(ContentFilter, self).__init__(imap, filterDef, globalConfig)

    @property
    def needsBody(self):
        """
        Get body requirements

        :return: True if filter needs mail BODY
        :rtype: bool
        """
        return True

class BodyFilter(ContentFilter):
    """
    Content filter

    .. seealso:: Filter
    """
    def __init__(self, imap, filterDef, globalConfig):
        """
        Create ContentFilter object

        :param imap: IMAP client
        :type imap: imapclient.IMAPClient
        :param filterDef: filter definition
        :type filterDef: dict
        """
        self.filter = filterDef
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))
        super(BodyFilter, self).__init__(imap, filterDef, globalConfig)

    def check(self, msgid, folder, message):
        """
        Run filter

        :param msgid:
        :type msgid: int
        :param message:
        :type message: email.email
        :return: True if ok
        :rtype: bool
        """
        result = self.filterFunction.check(msgid, self.extract(message))
        if result:
            self.log.info('Body of message %d positive', msgid)
            self.spam(msgid, folder, message)
        return not result

class HeaderFilter(ContentFilter):
    """
    Header filter

    .. seealso:: Filter
    """
    def __init__(self, imap, filterDef, globalConfig):
        """
        Create ContentFilter object

        :param imap: IMAP client
        :type imap: imapclient.IMAPClient
        :param filterDef: filter definition
        :type filterDef: dict
        """
        self.filter = filterDef
        self.header = filterDef['part'].lower()
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))
        super(HeaderFilter, self).__init__(imap, filterDef, globalConfig)

    def check(self, msgid, folder, message):
        """
        Run filter

        :param msgid:
        :type msgid: int
        :param message:
        :type message: email.email
        :return: True if ok
        :rtype: bool
        """
        field = DecodeHeader(message[self.header])
        result = self.filterFunction.check(msgid, field)
        self.log.debug('result: %s [%s]', result, field)
        if result:
            self.spam(msgid, folder, message)
        return not result

class PyzorFilter(Filter):
    """
    Filter with pyzor
    """

    def __init__(self, imap, filterDef, globalConfig):
        """
        Create a PyzorFilter object
        """
        self.filter = filterDef
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))
        self.pyzorClient = pyzor.client.Client()
        super(PyzorFilter, self).__init__(imap, filterDef, globalConfig)

    def check(self, msgid, folder, message):
        """
        Run filter

        :param msgid:
        :type msgid: int
        :param message:
        :type message: email.email
        :return: True if ok
        :rtype: bool
        """
        digest = DataDigester(message)
        presult = self.pyzorClient.check(digest.digest.hexdigest())
        if presult['Diag'] != 'OK':
            self.log.warning('Pyzor: %s ==> %s', str(presult).replace('\n', ' '))
        result = int(presult['Count']) == 0
        if not result:
            self.spam(msgid, folder, message)
        return result

    @property
    def needsBody(self):
        """
        Get body requirements

        :return: True if filter needs mail BODY
        :rtype: bool
        """
        return True

class BayesFilter(ContentFilter):
    """
    Bayes filter
    """

    def __init__(self, imap, filterDef, globalConfig):
        """
        Create a BayesFilter object
        """
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))
        super(BayesFilter, self).__init__(imap, filterDef, globalConfig)

    def check(self, msgid, folder, message):
        return False

class UrlFilter(Filter):
    """
    URL content filter
    """

    def __init__(self, imap, filterDef, globalConfig):
        """
        Create a UrlFilter object
        """
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))
        super(UrlFilter, self).__init__(imap, filterDef, globalConfig)

    def check(self, msgid, folder, message):
        """
        Run filter

        :param msgid:
        :type msgid: int
        :param message:
        :type message: email.email
        :return: True if ok
        :rtype: bool
        """
        # Check all parts
        if message.is_multipart():
            result = any(self.checkPart(m) for m in message.get_payload())
        else:
            result = self.checkPart(message)
        if result:
            self.spam(msgid, folder, message)
        return not result

    def checkPart(self, part):
        """
        Check part of e-mail for spam URLs

        :param part: payload
        :type part: email.message.Message
        :return: True if a spam URL was found
        :rtype: bool

        HTML has tags with href checked on top of text
        """
        payload = GetPayload(part)
        if part.get_content_type() == 'text/html':
            # First check href in tags
            soup = bs4.BeautifulSoup(payload, 'html.parser')
            for e in soup.findAll(href=True):
                parsed = urlparse(e['href'])
                if parsed.scheme:
                    url = parsed.netloc
                else:
                    url = parsed.path
                if CheckSpamUrl(url):
                    return True
            return False
        elif part.get_content_type() == 'text/plain':
            text = payload
            urls = set(m.group(1) for m in re.finditer(r'https?://([^/ \t\r\n">?,;:[\]{}!]+)', text))
            self.log.log(DEBUG2, 'URLs found: %s', urls)
            return any(CheckSpamUrl(u) for u in urls)



"""
Test body filter
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import *
import socket
from nose.tools import ok_, assert_raises
import email
import mock
from mailfilter import filters

def test10_positive_exact():
    """Positive for RBL"""
    imap = mock.Mock()
    tfilter = filters.RblFilter(imap, {'filter': 'rbl', 'org': 'spamcheck.swqzx', 'action': 'delete'}, {})
    message = email.message_from_file(open('test/rbl/spam.msg'))
    # Pretend gethostbyname HAS an entry about the IP in the mail
    with mock.patch('mailfilter.filters.gethostbyname', return_value='127.0.0.10') as f:
        ok_(not tfilter.check(1, '', message), 'RblFilter.check() failed')
        ok_(f.called, 'gethostbyname() never called')

def test20_negative_exact():
    """Negative for RBL"""
    imap = mock.Mock()
    tfilter = filters.RblFilter(imap, {'filter': 'rbl', 'org': 'spamcheck.plhb', 'action': 'delete'}, {})
    message = email.message_from_file(open('test/rbl/spam.msg'))
    # Pretend gethostbyname has no entry about the IP in the mail
    with mock.patch('mailfilter.filters.gethostbyname', side_effect=socket.gaierror) as f:
        ok_(tfilter.check(1, '', message), 'RblFilter.check() failed')
        ok_(f.called, 'gethostbyname() never called')



#!/usr/bin/env python

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division

from functools import partial

from future import standard_library
from jw.util.python3hell import SetDefaultEncoding

standard_library.install_aliases()
from builtins import *
import argparse

from . import __version__
import io
import logging
import logging.config
from jw.util import configuration
from .app import App
from gevent.monkey import patch_all
from .logging2 import *

VERSION = 'Mailfilter version {}'.format(__version__)

logging.addLevelName(DEBUG2, 'DEBUG2')
logging.addLevelName(DEBUG3, 'DEBUG3')

patch_all()

SetDefaultEncoding()

def Main():
    argp = argparse.ArgumentParser(description='IMAP Mail Account')
    argp.add_argument(
        '-c',
        '--config',
        action='store',
        type=partial(io.open, encoding='utf-8'),
        help='Specify configuration file to use'
    )
    argp.add_argument('-V', '--version', action='version', version=VERSION)
    args = argp.parse_args()
    config = configuration.FromStream(args.config)
    logging.config.dictConfig(config('logging'))
    App.init(config)
    App.run()
    App.terminate()

if __name__ == '__main__':
    Main()




# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Application
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from future import standard_library
import gevent
from six import PY2

from .folder import Folder

standard_library.install_aliases()
from builtins import *
from builtins import str
from builtins import object
import logging
from gevent import pool
import signal

#: Application name
Name = 'mailfilter'
#: Signal names
Signals = {v: k for k, v in list(signal.__dict__.items()) if k.startswith('SIG') and isinstance(v, int)}

DEFAULT_FOLDERS = ['INBOX']

class Application(object):
    """
    Application
    """
    def __init__(self):
        self.folders = []
        self.config = None
        self.greenletPool = None
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))
        self.log.info('created')

    def init(self, config):
        """
        Create Application

        :param config: configuration
        :type config: dict
        """
        self.log.info('init ==============================================================')
        self.config = config
        signal.signal(signal.SIGINT, Signal)
        signal.signal(signal.SIGTERM, Signal)
        signal.signal(signal.SIGQUIT, Signal)

    def run(self):
        """
        Run application

        """
        from .folder import Folder

        self.log.info('run')
        self.greenletPool = pool.Pool()
        for name, config in list(self.config('accounts').items()):
            if config.get('enabled', True):
                self.greenletPool.spawn(self.runFolder, name, config.folders, config, self.config('global', default={}))
            else:
                self.log.info('%s disabled', name)
        self.greenletPool.join()

    def runFolder(self, *args):
        """
        Run folder

        :param name:
        :type name:
        :param folders:
        :type folders:
        :param config:
        :type config:
        :param globalConfig:
        :type globalConfig:
        :return:
        :rtype:
        """
        af = Folder(*args)
        self.folders.append(af)
        af.run()

    def stop(self):
        self.greenletPool.kill(block=False)

    def terminate(self):
        """
        Terminate application

        """
        self.log.info('Terminating')

def Signal(signo, frame):
    """
    Stop application

    :param signo: signal number
    :type signo: int
    :param frame:
    """
    log = logging.getLogger('{}.{}'.format(Name, __name__))
    log.info('Got %s', Signals[signo])
    App.stop()

App = Application()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Part extraction functions
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging

from future import standard_library

standard_library.install_aliases()
from builtins import *
from builtins import str
from builtins import object

class Part(object):
    """
    Part base class
    """

class Body(Part):
    """
    Extract body from message

    All body parts are extracted, decoded and returned as one string
    """

    def __init__(self, filterDef):
        """
        Create Body object

        :param filterDef: filter definition
        :type filterDef: dict
        """
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))

    def __call__(self, message):
        """
        Get message part

        :return: message part
        :rtype: str
        """
        return self.extract(message)

    @staticmethod
    def extract(message):
        """
        Extract payload

        :param message:
        :type message:
        :return:
        :rtype:
        """
        if message.is_multipart():
            return ''.join(Body.extract(m) for m in message.get_payload())
        else:
            return message.get_payload(decode=True) or message.get_payload()

PART_FUNCTIONS = {
    'body': Body
}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Utilities
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals, print_function

from future import standard_library

from .header import decode_header

standard_library.install_aliases()
from builtins import *
import re

FromRe = re.compile(
    r'from\s+(?:'
    r'.*\(.*\[(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\]\)'
    '|'
    r'.*?\(.*?\)\s*?\((\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\)'
    r').*'
)

def D(prompt, value):
    print('{}: {}'.format(prompt, value))
    return value

def Flatten(items):
    """
    Flatten a list

    :param items:
    :type items: list
    :return: flattened list
    :rtype: iter
    """
    if isinstance(items, (list, tuple)):
        for item in items:
            if isinstance(item, (list, tuple)):
                for subitem in Flatten(item):
                    yield subitem
            else:
                yield item
    else:
        yield items

def DecodeHeader(header):
    try:
        d = decode_header(header)
        return ''.join(p[0].decode(p[1] or 'ascii', 'replace') for p in d)
    except Exception as e:
        return header



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Actions
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division

from functools import partial

from future import standard_library
from gevent import spawn
from jw.util.python3hell import Bytes2Str
from jw.util.retry import retry

standard_library.install_aliases()
from builtins import *
from email.mime.message import MIMEMessage
import imaplib
from itertools import chain
from smtplib import SMTPException, SMTP_SSL, SMTP
from email.mime.multipart import MIMEMultipart
import time
from urllib.request import urlopen

import pyzor

from pyzor.digest import DataDigester

from .util import FromRe
import logging
from .logging2 import *

Logger = logging.getLogger(__name__)

class Action(object):
    def __init__(self, imap, config, globalConfig):
        """
        Create Action object

        :param imap: IMAP client
        :type imap: imapclient.IMAPClient
        :param config_:
        :type config: dict

        Needs to be called from derived classes
        """
        self.imap = imap
        self.config = config
        self.globalConfig = globalConfig
        self._needsBody = False

    def run(self, msgid, folder, message):
        """
        Run filter

        :param int msgid: message ID
        :param str message: mail message

        To be derived in subclass
        """

    @property
    def needsBody(self):
        return self._needsBody

class CopyToAction(Action):
    """
    Action: copy to folder
    """
    def __init__(self, imap, config, globalConfig):
        """

        :param imap: IMAP client
        :type imap: imapclient.IMAPClient
        :param config:
        :type config: dict
        """
        self.folder = config[0]
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))
        super(CopyToAction, self).__init__(imap, config, globalConfig)

    def copy(self, msgid, separator='/'):
        """
        Copy message to another folder

        :param msgid: message ID
        :type msgid: int
        """
        # Make namespace-prefixed folder name (for Courier servers)
        personal = self.imap('namespace').personal[0]
        parent = personal[0].strip(personal[1])
        if parent:
            parent += personal[1]
        folder = parent + self.folder.replace('/', separator)
        # Copy message
        for _ in range(2):
            try:
                self.imap.copy(msgid, folder)
                break
            except imaplib.IMAP4.error:
                # On IMAP error assume destination folder does not exist. Create it.
                try:
                    self.imap.create_folder(folder)
                except Exception as e:
                    self.log.critical('Could not create folder: %s', e)
                else:
                    self.log.info('%s: Created folder %s', self.imap.host, folder)
            except Exception as e:
                self.log.critical('Unrecoverable error from imap.copy(): %s', e)
                return False
        else:
            self.log.critical('Failed to copy message %s', msgid)
            return False
        return True

    def run(self, msgid, folder, message):
        """
        Run action

        :param msgid: Message ID
        :type msgid: int
        :param message: Message object
        :type message: email.email
        """
        if self.copy(msgid, folder.separator):
            self.log.info('Message %d copied to folder %s', msgid, self.folder)

class MoveToAction(CopyToAction):
    def __init__(self, imap, config, globalConfig):
        """
        Create MoveToAction object

        :param imap: IMAP client
        :type imap: imapclient.IMAPClient
        :param config:
        :type config: dict
        """
        super(MoveToAction, self).__init__(imap, config, globalConfig)

    def run(self, msgid, folder, message):
        """

        :param msgid:
        :param message:
        """
        if self.copy(msgid, folder.separator):
            self.imap.delete_messages(msgid)
            self.imap.expunge()
            self.log.info('Message %d moved to folder %s', msgid, self.folder)
        else:
            self.log.critical('Copy failed, could not move message %s', msgid)

class DeleteAction(Action):
    """
    Action: delete message
    """

    def __init__(self, imap, config, globalConfig):
        """
        Create DeleteAction object

        :param imap: IMAP client
        :type imap: imapclient.IMAPClient
        :param config:
        :type config: dict
        """
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))
        super(DeleteAction, self).__init__(imap, config, globalConfig)

    def run(self, msgid, folder, message):
        """

        :param msgid:
        :param message:
        """
        self.imap.delete_messages(msgid)
        self.imap.expunge()
        self.log.info('Message %d deleted', msgid)

class SetFlagAction(Action):
    """
    Action: set a flag
    """

    def __init__(self, imap, config, globalConfig):
        """
        Create SetFlagAction object

        :param imap: IMAP client
        :type imap: imapclient.IMAPClient
        :param config:
        :type config: dict
        """
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))
        self.flags = config
        super(SetFlagAction, self).__init__(imap, config, globalConfig)

    def run(self, msgid, folder, message):
        """

        :param msgid:
        :param message:
        """
        self.imap.set_flags(msgid, self.flags)
        self.log.info('Message %d flag(s) %s set', msgid, self.flags)

class ClearFlagAction(Action):
    """
    Action: clear a flag
    """

    def __init__(self, imap, config, globalConfig):
        """
        Create DeleteAction object

        :param imap: IMAP client
        :type imap: imapclient.IMAPClient
        :param config:
        :type config: dict
        """
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))
        self.flags = config
        super(ClearFlagAction, self).__init__(imap, config, globalConfig)

    def run(self, msgid, folder, message_):
        """

        :param msgid:
        :param message_:
        """
        self.imap.remove_flags(msgid, self.flags)
        self.log.info('Message %d flag(s) %s cleared', msgid, self.flags)

class PyzorReportAction(Action):
    """
    Action: report spam to Pyzor
    """

    def __init__(self, imap, config, globalConfig):
        """
        Create a PyzorReport object
        """
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))
        self.pyzorClient = pyzor.client.Client()
        super(PyzorReportAction, self).__init__(imap, config, globalConfig)

    @property
    def needsBody(self):
        return True

    def run(self, msgid, folder, message):
        """

        :param msgid:
        :param message:
        """
        digest = DataDigester(message)
        try:
            hexdigest = digest.digest.hexdigest()
            self.pyzorClient.report(hexdigest)
        except Exception as e:
            self.log.critical('From self.pyzorClient.report("%s"):', hexdigest, e)
        else:
            self.log.info('Reported %s to pyzor (digest %s)', msgid, hexdigest)

class ForwardAction(Action):
    """
    Action: forward mail
    """

    def __init__(self, imap, config, globalConfig):
        """
        Create a ForwardAction object
        """
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))
        super(ForwardAction, self).__init__(imap, config, globalConfig)

    def run(self, msgid, folder, message):
        """

        :param int msgid:
        :param str message:
        """
        fmessage = MIMEMultipart()
        smtp = folder.config.get('smtp', self.globalConfig['smtp'])
        username = smtp.get('username', folder.config['username'])
        password = smtp.get('password', folder.config['password'])
        fmessage['From'] = smtp.get('from', username)
        fmessage['To'] = self.config[0]
        fmessage['Date'] = time.ctime()
        fmessage['Subject'] = 'Fwd: %s' % Bytes2Str(message['subject'])
        fmessage.attach(MIMEMessage(message))
        spawn(retry, self.send, fmessage, msgid, password, smtp, username)

    def send(self, fmessage, msgid, password, smtp, username):
        self.log.debug('Forwarding message %s', msgid)
        try:
            session = SMTP(smtp['host'])
            session.starttls()
            self.log.log(DEBUG2, 'STARTTLS connection to %s established', smtp['host'])
        except SMTPException as e:
            session = SMTP_SSL(smtp['host'])
            self.log.log(DEBUG2, 'SSL connection to %s established', smtp['host'])
        session.login(username, password)
        session.sendmail(fmessage['from'], self.config[0], fmessage.as_string())
        session.quit()
        self.log.info('Message %s forwarded to %s', msgid, self.config[0])

class StopAction(Action):
    """
    Action: stop action sequence
    """

    def __init__(self, imap, config, globalConfig):
        """
        Create StopAction object

        :param imap: IMAP client
        :type imap: imapclient.IMAPClient
        :param config:
        :type config: dict
        """
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))
        super(StopAction, self).__init__(imap, config, globalConfig)

    def run(self, msgid, folder, message):
        """

        :param msgid:
        :param message:
        """
        self.log.info('%s: Stop', msgid)
        return 'stop'

class ReportBadipsAction(Action):
    """
    Action: report to badips.com
    """

    def __init__(self, imap, config, globalConfig):
        """
        Create ReportBadipsAction object

        :param imap: IMAP client
        :type imap: imapclient.IMAPClient
        :param config:
        :type config: dict
        """
        self.log = logging.getLogger('{}.{}'.format(str(self.__module__), self.__class__.__name__))
        super(ReportBadipsAction, self).__init__(imap, config, globalConfig)

    def run(self, msgid, folder, message):
        """

        :param msgid:
        :param message:
        """
        received = message.get_all('received')
        if received:
            try:
                mta = next(
                    # First non-empty element of tuple
                    t[0] or t[1]
                        # Tuples with one of the elements containing an IP
                        for t in chain.from_iterable(FromRe.findall(i) for i in received if i.startswith('from '))
                )
            except StopIteration:
                self.log.warning('No "Received: from" header in message %s', msgid)
                return
            try:
                url = 'https://www.badips.com/add/%s/%s' % ('badbots', mta)
                self.log.log(DEBUG2, 'HTTP GET %s', url)
                self.log.info("%s: Reported %s -> %s", msgid, mta, urlopen(url).read())
            except Exception as e:
                self.log.critical('%s: Reporting spammer %s as "badbots": %s', msgid, mta, e)
        else:
            self.log.warning('No Received header in message %s', msgid)



#!/usr/bin/env python

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup, find_packages

setup(
    name="jw.mailfilter",
    version="0.5",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'setuptools',
        'jw.util',
        'gevent',
        'dnspython',
        'imapclient',
        'pyzor',
        'beautifulsoup4',
        'future',
        'sphinxjp.themes.dotted',
    ],
    package_data={
        '': ['*.rst', '*.txt']
    },
    entry_points={
        'console_scripts': [
            'mailfilter = mailfilter.main:Main'
        ],
        'jw.mailfilter.filter': [
            'all = mailfilter.filters:All',
            'rbl = mailfilter.filters:RblFilter',
            'pyzor = mailfilter.filters:PyzorFilter',
            'url = mailfilter.filters:UrlFilter',
            'body = mailfilter.filters:BodyFilter',
            'header = mailfilter.filters:HeaderFilter',
        ],
        'jw.mailfilter.action': [
            'copy-to = mailfilter.action:CopyToAction',
            'move-to = mailfilter.action:MoveToAction',
            'delete = mailfilter.action:DeleteAction',
            'set-flag = mailfilter.action:SetFlagAction',
            'clear-flag = mailfilter.action:ClearFlagAction',
            'report-pyzor = mailfilter.action:PyzorReportAction',
            'report-badips = mailfilter.action:ReportBadipsAction',
            'forward-to = mailfilter.action:ForwardAction',
            'stop = mailfilter.action:StopAction',
        ]
    },
    test_suite='nose.collector',
    tests_require=['nose', 'mock'],
    author="Johnny Wezel",
    author_email="dev-jay@wezel.name",
    description="IMAP mail filter",
    long_description='IMAP mail filter',
    license="GPL",
    platforms='POSIX',
    keywords="mail, imap, filter, spam",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: No Input/Output (Daemon)',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
    ],
    url="https://pypi.python.org/pypi/jw.mailfilter"
)
import setuptools.command.install

